REBOL [
	Title: "REBOL/Services - Async-modes HTTP Client Handler (for View 1.3)"
	Author: "Gabriele Santilli"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- HTTP Client Handler for non-/async Core ----------------------------------
;
; NOTE:
;   This code uses HTTP/1.0. It does not really need the chunked
;   transfer mode of HTTP/1.1. We can do this because the length of
;   a RIS packet is encoded in its header near the beginning of the
;   transfer. The LOAD-PACKET function scans for the RIS header get
;   the length, so no HTTP content length info is needed.

add-client-handler [

	scheme: 'http

	wait-find: func [port [port!]][find system/ports/wait-list port]
	wait-stop: func [port [port!]][remove wait-find port]
	wait-start: func [port [port!]][insert tail system/ports/wait-list port]

	opener: func [
		spec [port!]
		config [object! none!]
	][
		debug connect ['opening-http-port spec/host spec/port-id]
		if config [
			; Build new TCP port based on HTTP port.
			; Let the original HTTP port get recycled.
			spec/handler: none ;spec/proxy: none ; help out GC
			spec: make port! [
				scheme: 'tcp
				proxy: spec/proxy spec/proxy: none
				host: any [proxy/host spec/host]
				port-id: any [proxy/port-id spec/port-id]
				path: spec/path
				target: spec/target
				url: join http:// [spec/host #":" spec/port-id #"/"]
			]
			; Create a session and open the port:
			make-session spec config
			spec/locals: make object! [
				busy?:          false   ; is true when we are inside on-data, used to avoid recursion
				max-retry:      25      ; max number of connection attempts (none! = forever) -Romano
				retry:          0       ; counter for retry -Romano
				max-transfer:   16 * 1024  ; max TCP transfer size accepted
				;dnsport:        none    ; async dns port
				wakeup?:        none    ; if true, returns from wait
				outBuffer:      make binary! 1024
				first-line:     none    ; first line of response (to check HTTP response code)
			]
			; !!! does not work in many cases (i.e. connect)
			spec/timeout: spec/user-data/port-timeout
			wait-start spec
		]
		; reset status (if we are reopening the port)
		spec/locals/first-line: none
		spec/locals/busy?: false
		spec/locals/retry: 0
		spec/locals/wakeup?: none
		clear spec/locals/outBuffer
		spec/async-modes: 'connect
		spec/awake: :on-connect
		spec/status: 'close
		on-connect spec
		spec ; not yet open at this point!
	]

	no-block:       -3      ; return code from IO indicating not to block
	would-block?:   func [err] [517 = get in err 'code] ; checks blocking error

	on-connect: func [port /local err modes] [
		either error? err: try [
			open/direct/binary/no-wait port
		] [
			if would-block? disarm err [
				if all [
					;check the number of retry -Romano
					integer? port/locals/max-retry
					port/locals/max-retry < port/locals/retry: port/locals/retry + 1
				][
					print ["*** Client Error: Cannot connect to" port/host] 
					attempt [close port]
					wait-stop port
				]
				return false
			]
			print ["*** Client Error: " mold disarm err] 
			attempt [close port]
			wait-stop port
			return false
		] [
			port/awake: :on-data
			; REBOL should do it automatically
			;set-modes port [async-modes: 'read]
			debug connect ['open port/remote-ip port/remote-port port/local-port]
			port/status: 'open
			send-request port
		]
		false
	]

	on-data: func [port /local data err res max-transfer outBuffer line wakeup?] [
		;debug async/locals/busy? "Recursive call to on-data!!! (Or, most likely, it was interrupted last time.)"
		max-transfer: port/locals/max-transfer
		res: false
		port/locals/busy?: true
		; try to receive
		data: make string! max-transfer + 2
		wakeup?: found? port/locals/wakeup?
		either error? err: try [read-io port data max-transfer] [
			if not would-block? disarm :err [
				print ["*** Client Error: " mold disarm err] 
				attempt [close port]
				wait-stop port
				port/locals/busy?: false
				return wakeup?
			]
		] [
			if all [err < -1 err <> no-block] [
				print ["*** Client Error: read-io aborted (" err ")"] 
				attempt [close port]
				wait-stop port
				port/locals/busy?: false
				return wakeup?
			]
			if find [0 -1] err [
				; port closed
				debug connect ['close port/user-data/id port/remote-ip port/remote-port port/local-port]
				attempt [close port]
				port/status: 'close
				; Check to see if there are any other pending requests:
				if not empty? port/user-data/send-queue [
					;reset-async-port port   doesn't seem to be needed in 1.3,
					;                        however, i guess we have a problem if the
					;                        close above did not work.
					;port/user-data/config/handler-context/opener port none  why not using opener directly?
					opener port none
				]
				port/locals/busy?: false
				return found? port/locals/wakeup?  ; not local wakeup? because could be changed
			]
			if err > 0 [
				append port/user-data/recv-buf data
				debug tcp-read [err 'bytes length? port/user-data/recv-buf 'total]
				; get first line of response
				if not port/locals/first-line [
					port/locals/first-line: copy-line as-string port/user-data/recv-buf
				]
				while [ ; multiple requests may be concatenated in the buffer
					all [
						not empty? port/user-data/recv-buf
						data: load-packet port/user-data
					]
				][
					wake-request port data ; arg is block (ok) or word (failure)
				] 
				; Deal with possible HTTP server errors...
				if not data [ ; No packet, what's up?
					line: port/locals/first-line ; server reply
					if not find line "200" [ ; if not 200, then problem
						debug error ['bad-cgi line] ; server reply
						;print as-string session/recv-buf ; !!!
						clear port/user-data/recv-buf
						wake-request port 'http-server-error
					]
				]
				;res: found? port/locals/wakeup?  actually, better to do it on close
			]
		]
		; try to send
		outBuffer: port/locals/outBuffer
		if not empty? outBuffer [
			either error? err: try [
				write-io port outBuffer min max-transfer length? outBuffer
			] [
				if not would-block? disarm :err [
					print ["*** Client Error: " mold disarm err] 
					attempt [close port]
					wait-stop port
					port/locals/busy?: false
					return false
				]
			] [
				if all [err < -1 err <> no-block] [
					print ["*** Client Error: write-io aborted (" err ")"] 
					attempt [close port]
					wait-stop port
					port/locals/busy?: false
					return false
				]
				if err > 0 [
					port/locals/outBuffer: outBuffer: skip outBuffer err
					if tail? outBuffer [
						port/locals/outBuffer: clear head outBuffer
						; Top item in queue is always the one just sent
						if data: port/user-data/send-queue/1 [append port/user-data/sent-queue data]
						remove port/user-data/send-queue
						;send-request port    hmm, really?
						;res: res or found? port/locals/wakeup?
						;if no more data to write change the mode to 'read
						;if empty? port/locals/outBuffer [port/async-modes: 'read]
						port/async-modes: 'read
					]
				]
			]
		]
		port/locals/busy?: false
		res
	]

	sender: func [
		port [port!]
		request
	][
		request: mold-packet port/user-data request
		make-http-request port request
		insert tail port/locals/outBuffer request
		set-modes port [async-modes: [read write]]
	]

	set-wakeup: func [
		port [port!]
		flag [logic!]
	] [
		port/locals/wakeup?: any [port/locals/wakeup? flag]
	]

	clear-wakeup: func [
		port [port!]
	] [
		port/locals/wakeup?: none
	]

	form-http-host: func [
		"Create host name with port number if not the default."
		port
		/local host id
	][
		; FORM instead of COPY because host can be a tuple!
		host: form port/host
		if all [id: port/port-id id <> 80] [repend host [#":" id]]
		host
	]

	mold-http-target: func [
		"Create path and target on host server (escaped if needed)."
		port
		/local target
	][
		target: either port/proxy/host [copy port/url] [to-url "/"]
		if port/path [append target port/path]
		if port/target [append target port/target]
		mold target ; escape chars, return result
	]

	make-http-request: func [
		"Add an HTTP header to the request packet."
		port
		data "The request packet - modified by this function"
	][
		head insert data reduce [
			"POST " mold-http-target port " HTTP/1.0" crlf
			"Host: " form-http-host port crlf
			"Cache-Control: no-cache" crlf
			"Pragma: no-cache" crlf
			"Content-Length: " length? data crlf
			; keep-alive? !!!
			crlf
		]
	]

]
