REBOL [
	Title: "REBOL/Services - Async-modes TCP Client Handler (for View 1.3)"
	Author: "Gabriele Santilli"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- TCP Client Handler for non-/async Core -----------------------------------

; uses code from async-protocol.r by Gabriele Santilli, Romano Paolo Tenca and
; Maarten Koopmans

; NOTE: the port is always added to the system wait-list.
; Also: READ protocol error might blow it out to console.

add-client-handler [

	scheme: 'tcp

	wait-find: func [port [port!]][find system/ports/wait-list port]
	wait-stop: func [port [port!]][remove wait-find port]
	wait-start: func [port [port!]][insert tail system/ports/wait-list port]

	opener: func [
		spec [port!]
		config [object!]
	][
		debug connect ['opening-tcp-port spec/host spec/port-id]
		; Create a session and open the port:
		make-session spec config
		if not spec/port-id [spec/port-id: rebol-port-id]
		spec/locals: make object! [
			busy?:          false   ; is true when we are inside on-data, used to avoid recursion
			max-retry:      5       ; max number of connection attempts (none! = forever) -Romano
			retry:          0       ; counter for retry -Romano
			max-transfer:   16 * 1024  ; max TCP transfer size accepted
			;dnsport:        none    ; async dns port
			wakeup?:        none    ; if true, returns from wait
			outBuffer:      make binary! 1024
		]
		spec/async-modes: 'connect
		spec/awake: :on-connect
		spec/status: none
		; !!! does not work in many cases (i.e. connect)
		spec/timeout: spec/user-data/port-timeout
		wait-start spec
		on-connect spec
		spec ; not yet open at this point!
	]

	no-block:       -3      ; return code from IO indicating not to block
	would-block?:   func [err] [517 = get in err 'code] ; checks blocking error

	on-connect: func [port /local err modes] [
		;print "ON-CONNECT"
		;print [port/host port/port-id port/async-modes found? find system/ports/wait-list port]
		either error? err: try [
			open/direct/binary/no-wait port
		] [
			if would-block? disarm err [
				if all [
					;check the number of retry -Romano
					integer? port/locals/max-retry
					port/locals/max-retry < port/locals/retry: port/locals/retry + 1
				][
					print ["*** Client Error: Cannot connect to" port/host] 
					attempt [close port]
					wait-stop port
				]
				return false
			]
			print ["*** Client Error: " mold disarm err] 
			attempt [close port]
			wait-stop port
			return false
		] [
			port/awake: :on-data
			; REBOL should do it automatically
			;set-modes port [async-modes: 'read]
			debug connect ['open port/remote-ip port/remote-port port/local-port]
			port/status: 'open
			send-request port
		]
		false
	]

	on-data: func [port /local data err res max-transfer outBuffer] [
		;debug async/locals/busy? "Recursive call to on-data!!! (Or, most likely, it was interrupted last time.)"
		max-transfer: port/locals/max-transfer
		res: false
		port/locals/busy?: true
		; try to receive
		data: make string! max-transfer + 2
		either error? err: try [read-io port data max-transfer] [
			if not would-block? disarm :err [
				print ["*** Client Error: " mold disarm err] 
				attempt [close port]
				wait-stop port
				port/locals/busy?: false
				return false
			]
		] [
			if all [err < -1 err <> no-block] [
				print ["*** Client Error: read-io aborted (" err ")"] 
				attempt [close port]
				wait-stop port
				port/locals/busy?: false
				return false
			]
			if find [0 -1] err [
				; port closed
				attempt [close port]
				port/status: none
				port/locals/busy?: false
				return false
			]
			if err > 0 [
				append port/user-data/recv-buf data
				debug tcp-read [err 'bytes length? port/user-data/recv-buf 'total]
				while [ ; multiple requests may be concatenated in the buffer
					all [
						not empty? port/user-data/recv-buf
						data: load-packet port/user-data
					]
				][wake-request port data]
				res: found? port/locals/wakeup?
			]
		]
		; try to send
		outBuffer: port/locals/outBuffer
		if not empty? outBuffer [
			either error? err: try [
				write-io port outBuffer min max-transfer length? outBuffer
			] [
				if not would-block? disarm :err [
					print ["*** Client Error: " mold disarm err] 
					attempt [close port]
					wait-stop port
					port/locals/busy?: false
					return false
				]
			] [
				if all [err < -1 err <> no-block] [
					print ["*** Client Error: write-io aborted (" err ")"] 
					attempt [close port]
					wait-stop port
					port/locals/busy?: false
					return false
				]
				if err > 0 [
					port/locals/outBuffer: outBuffer: skip outBuffer err
					if tail? outBuffer [
						port/locals/outBuffer: clear head outBuffer
						; Top item in queue is always the one just sent
						if data: port/user-data/send-queue/1 [append port/user-data/sent-queue data]
						remove port/user-data/send-queue
						send-request port
						res: res or found? port/locals/wakeup?
						;if no more data to write change the mode to 'read
						if empty? port/locals/outBuffer [port/async-modes: 'read]
					]
				]
			]
		]
		port/locals/busy?: false
		res
	]

	sender: func [
		port [port!]
		request
	][
		;print "SENDER"
		;print [port/host port/port-id port/async-modes]
		insert tail port/locals/outBuffer mold-packet port/user-data request
		set-modes port [async-modes: [read write]]
	]

	set-wakeup: func [
		port [port!]
		flag [logic!]
	] [
		port/locals/wakeup?: any [port/locals/wakeup? flag]
	]

	clear-wakeup: func [
		port [port!]
	] [
		port/locals/wakeup?: none
	]
]
